# -*- coding: utf8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .solution import Solution
import numpy as np

class Nipk(Solution):
    """Non-iterative p-k method for flutter solution
    """
    def __init__(self, name, flu, eig, agg, mach, k_ref, l_ref, vrb, n_modes, n_speed):
        super().__init__(name, flu, eig, agg, mach, k_ref, l_ref, vrb, n_modes, n_speed)
        self.name = 'NIPK' # solver name
        self._pk = np.zeros((self.n_speed, self.n_modes, 2), dtype=complex) # eigenvalues (at reference reduced frequencies)
        self._vk = np.zeros((self.n_speed, self.n_modes, 2, self.n_modes), dtype=complex) # eigenmodes (at reference reduced frequencies)
        self._ik = np.zeros((self.n_speed, self.n_modes, 2), dtype=int) # indices of reduced frequencies required for interpolation
        self._kk = np.zeros((self.n_speed, self.n_modes, 2)) # reduced frequencies required for interpolation
        self._dk = np.zeros((self.n_speed, self.n_modes, 2)) # difference between eigenvalues and frequencies (at reference reduced frequencies)
        self._p = np.zeros((self.n_speed, self.n_modes), dtype=complex) # interpolated eigenvalue

    def compute(self):
        """Compute frequencies and dampings
        """
        self.rho, self.speed = self._flu.compute(self._mach)
        for ivel in range(self.n_speed):
            # Compute eigenvalues at zero airspeed
            if self.speed[ivel] == 0:
                for imod in range(self.n_modes):
                    self.freq[ivel, imod] = 0.
                continue
            # Compute eigenvalues for each reference reduced frequency
            p_k = np.zeros((self.n_freqs, self.n_modes), dtype=complex)
            v_k = np.zeros((self.n_freqs, self.n_modes, self.n_modes), dtype=complex)
            for ifrq in range(self.n_freqs):
                p_k[ifrq,:], v_k[ifrq,:,:] = self._eig.compute(self.rho[ivel], self.speed[ivel], self._m, self._k, self._q[ifrq,:,:])
            # Match reduced frequency by linear interpolation for each mode
            for imod in range(self.n_modes):
                delta, idx = self.__find_index(self.speed[ivel], p_k[:, imod], self._kref)
                p_re, p_im = self.__interp(self._kref[idx], delta[idx], p_k[idx, imod])
                # Compute frequency and damping
                self.freq[ivel, imod] = p_im * self.speed[ivel] / self._lref / (2 * np.pi)
                self.damp[ivel, imod] = p_re / p_im
                # Save intermediate results
                self._ik[ivel, imod, :] = idx
                self._kk[ivel, imod, :] = self._kref[idx]
                self._dk[ivel, imod, :] = delta[idx]
                self._pk[ivel, imod, :] = p_k[idx, imod]
                self._vk[ivel, imod, :, :] = v_k[idx, :, imod]
                self._p[ivel, imod] = p_re + 1j * p_im
        # Compute aggregated damping over modes then over dynamic pressures
        self._ksm = np.zeros(self.n_speed)
        for i in range(self.n_speed):
            self._ksm[i] = self._agg.compute(self.damp[i,:])
        self.damp_agg = self._agg.compute(self._ksm)

    def compute_gradients(self):
        """Compute the derivatives of the damping aggregations wrt mass, stiffness and aerodynamic terms
        """
        z = np.zeros((self.n_modes, self.n_modes))
        zk = np.zeros((self.n_freqs, self.n_modes, self.n_modes))
        # Derivative wrt M and K
        for i in range(z.size):
            dm = np.zeros((self.n_modes, self.n_modes))
            dm[np.unravel_index(i, dm.shape)] = 1.
            self.damp_m[i] = self.__compute_dg(dm, z, zk+1j*zk)
            dk = np.zeros((self.n_modes, self.n_modes))
            dk[np.unravel_index(i, dk.shape)] = 1.
            self.damp_k[i] = self.__compute_dg(z, dk, zk+1j*zk)
        # Derivative wrt Re(Q) and Im(Q)
        for i in range(zk.size):
            dq = np.zeros((self.n_freqs, self.n_modes, self.n_modes))
            dq[np.unravel_index(i, dq.shape)] = 1.
            self.damp_qre[i] = self.__compute_dg(z, z, dq+1j*zk)
            dq = np.zeros((self.n_freqs, self.n_modes, self.n_modes))
            dq[np.unravel_index(i, dq.shape)] = 1.
            self.damp_qim[i] = self.__compute_dg(z, z, zk+1j*dq)

    def __find_index(self, u, p, k):
        """Find the indices between which the frequency must be interpolated
        u : freestream velocity
        p : eigenvalue solutions
        k : reference reduced frequencies
        """
        # Compute the difference between the solution and the input
        delta = np.zeros(len(k))
        for i in range(len(k)):
            delta[i] = np.imag(p[i]) - k[i]
        # Find the indices between which the difference is zero
        for i in range(len(delta) - 1):
            if (delta[i] < 0 and delta[i+1] > 0) or (delta[i] > 0 and delta[i+1] < 0):
                idx = [i, i+1]
                break
            if i == len(delta) - 2:
                ai = np.argmin(np.abs(delta))
                if ai == 0:
                    idx = [ai, ai+1]
                elif ai == len(delta) - 1:
                    idx = [ai-1, ai]
                else:
                    raise RuntimeError(f'NIPK: could not match frequency for mode {i} at velocity {u}!\n')
        return delta, idx

    def __interp(self, k, delta, p):
        """Match the frequency and interpolate the eigenvalues and eigenvectors
        k : reference reduced frequencies
        delta : differences between solution and frequency
        p : eigenvalue solutions
        """
        # Interpolate the frequency at which the difference is zero
        slope = (delta[1] - delta[0]) / (k[1] - k[0])
        p_im = (slope * k[0] - delta[0]) / slope
        # Interpolate the damping
        slope = (np.real(p[1]) - np.real(p[0])) / (k[1] - k[0])
        p_re = np.real(p[0]) + slope * (p_im - k[0])
        return p_re, p_im

    def __compute_dg(self, dm, dk, dq_k):
        """Compute the derivative of the damping aggregation wrt one input
        m : derivative of mass matrix
        k : derivative of stiffness matrix
        q : derivative of aerodynamic matrix
        """
        dg = np.zeros((self.n_speed, self.n_modes))
        for ivel in range(self.n_speed):
            # Derivative of eigenvalue solution at zero airspeed
            if self.speed[ivel] == 0:
                continue
            # Derivative of eigenvalue solution
            dp_k = np.zeros((self.n_modes, 2), dtype=complex)
            for imod in range(self.n_modes):
                for ifrq in range(2):
                    q = self._q[self._ik[ivel, imod, ifrq], :, :]
                    p_k = self._pk[ivel, imod, ifrq]
                    v_k = self._vk[ivel, imod, ifrq, :].T
                    dq = dq_k[self._ik[ivel, imod, ifrq], :, :]
                    dp_k[imod, ifrq] = self._eig.compute_gradients(self.rho[ivel], self.speed[ivel], self._m, self._k, q, p_k, v_k, dm, dk, dq)
            for imod in range(self.n_modes):
                # Derivative of eigenvalue interpolation
                dp_re, dp_im = self.__dinterp(self._kk[ivel, imod, :], self._dk[ivel, imod, :], self._pk[ivel, imod, :], np.imag(self._p[ivel, imod]), dp_k[imod, :])
                # Derivative of damping
                dg[ivel, imod] = dp_re / np.imag(self._p[ivel, imod]) - np.real(self._p[ivel, imod]) / np.imag(self._p[ivel, imod])**2 * dp_im
        # Derivative of double aggregation
        dksm = np.zeros(self.n_speed)
        for ivel in range(self.n_speed):
            dksm[ivel] = self._agg.compute_gradients(self.damp[ivel, :], dg[ivel, :])
        return self._agg.compute_gradients(self._ksm, dksm)

    def __dinterp(self, k, delta, p, p_im, dp):
        """Compute the derivative of a linear interpolation
        k : reference reduced frequencies
        delta : differences between solution and frequency
        p : eigenvalue solutions
        p_im : imaginary part of eigenvalue solutions
        dp : derivatives of eigenvalue solution
        """
        # x: 0 = s * (x - x0) + y0
        slope = (delta[1] - delta[0]) / (k[1] - k[0])
        d_slope = (np.imag(dp[1]) - np.imag(dp[0])) / (k[1] - k[0])
        dp_im = (d_slope * k[0] - np.imag(dp[0])) / slope - (slope * k[0] - delta[0]) / slope**2 * d_slope
        # z: z = s * (x - x0) + z0
        slope = (np.real(p[1]) - np.real(p[0])) / (k[1] - k[0])
        d_slope = (np.real(dp[1]) - np.real(dp[0])) / (k[1] - k[0])
        dp_re = np.real(dp[0]) + d_slope * (p_im - k[0]) + slope * dp_im
        return dp_re, dp_im
